#!/usr/bin/env python3
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run, SubGroup
from caproto import ChannelType


class EpicsScalerGroup(PVGroup):
    count = pvproperty(name='.CNT', dtype=int)
    count_mode = pvproperty(value='OneShot',
                            name='.CONT',
                            dtype=ChannelType.ENUM,
                            enum_strings=['OneShot', 'AutoCount'])
    delay = pvproperty(name='.DLY', dtype=float)
    auto_count_delay = pvproperty(name='.DLY1', dtype=float)

    class ChannelsGroup(PVGroup):
        chan1 = pvproperty(value=0, name='.S1', dtype=int, read_only=True)
        chan2 = pvproperty(value=0, name='.S2', dtype=int, read_only=True)
        chan3 = pvproperty(value=0, name='.S3', dtype=int, read_only=True)
        chan4 = pvproperty(value=0, name='.S4', dtype=int, read_only=True)
        chan5 = pvproperty(value=0, name='.S5', dtype=int, read_only=True)
        chan6 = pvproperty(value=0, name='.S6', dtype=int, read_only=True)
        chan7 = pvproperty(value=0, name='.S7', dtype=int, read_only=True)
        chan8 = pvproperty(value=0, name='.S8', dtype=int, read_only=True)
        chan9 = pvproperty(value=0, name='.S9', dtype=int, read_only=True)
        chan10 = pvproperty(value=0, name='.S10', dtype=int, read_only=True)
        chan11 = pvproperty(value=0, name='.S11', dtype=int, read_only=True)
        chan12 = pvproperty(value=0, name='.S12', dtype=int, read_only=True)
        chan13 = pvproperty(value=0, name='.S13', dtype=int, read_only=True)
        chan14 = pvproperty(value=0, name='.S14', dtype=int, read_only=True)
        chan15 = pvproperty(value=0, name='.S15', dtype=int, read_only=True)
        chan16 = pvproperty(value=0, name='.S16', dtype=int, read_only=True)
        chan17 = pvproperty(value=0, name='.S17', dtype=int, read_only=True)
        chan18 = pvproperty(value=0, name='.S18', dtype=int, read_only=True)
        chan19 = pvproperty(value=0, name='.S19', dtype=int, read_only=True)
        chan20 = pvproperty(value=0, name='.S20', dtype=int, read_only=True)
        chan21 = pvproperty(value=0, name='.S21', dtype=int, read_only=True)
        chan22 = pvproperty(value=0, name='.S22', dtype=int, read_only=True)
        chan23 = pvproperty(value=0, name='.S23', dtype=int, read_only=True)
        chan24 = pvproperty(value=0, name='.S24', dtype=int, read_only=True)
        chan25 = pvproperty(value=0, name='.S25', dtype=int, read_only=True)
        chan26 = pvproperty(value=0, name='.S26', dtype=int, read_only=True)
        chan27 = pvproperty(value=0, name='.S27', dtype=int, read_only=True)
        chan28 = pvproperty(value=0, name='.S28', dtype=int, read_only=True)
        chan29 = pvproperty(value=0, name='.S29', dtype=int, read_only=True)
        chan30 = pvproperty(value=0, name='.S30', dtype=int, read_only=True)
        chan31 = pvproperty(value=0, name='.S31', dtype=int, read_only=True)
        chan32 = pvproperty(value=0, name='.S32', dtype=int, read_only=True)

    channels = SubGroup(ChannelsGroup, prefix='')

    class NamesGroup(PVGroup):
        name1 = pvproperty(value='name', name='.NM1', dtype=ChannelType.STRING)
        name2 = pvproperty(value='name', name='.NM2', dtype=ChannelType.STRING)
        name3 = pvproperty(value='name', name='.NM3', dtype=ChannelType.STRING)
        name4 = pvproperty(value='name', name='.NM4', dtype=ChannelType.STRING)
        name5 = pvproperty(value='name', name='.NM5', dtype=ChannelType.STRING)
        name6 = pvproperty(value='name', name='.NM6', dtype=ChannelType.STRING)
        name7 = pvproperty(value='name', name='.NM7', dtype=ChannelType.STRING)
        name8 = pvproperty(value='name', name='.NM8', dtype=ChannelType.STRING)
        name9 = pvproperty(value='name', name='.NM9', dtype=ChannelType.STRING)
        name10 = pvproperty(value='name', name='.NM10',
                            dtype=ChannelType.STRING)
        name11 = pvproperty(value='name', name='.NM11',
                            dtype=ChannelType.STRING)
        name12 = pvproperty(value='name', name='.NM12',
                            dtype=ChannelType.STRING)
        name13 = pvproperty(value='name', name='.NM13',
                            dtype=ChannelType.STRING)
        name14 = pvproperty(value='name', name='.NM14',
                            dtype=ChannelType.STRING)
        name15 = pvproperty(value='name', name='.NM15',
                            dtype=ChannelType.STRING)
        name16 = pvproperty(value='name', name='.NM16',
                            dtype=ChannelType.STRING)
        name17 = pvproperty(value='name', name='.NM17',
                            dtype=ChannelType.STRING)
        name18 = pvproperty(value='name', name='.NM18',
                            dtype=ChannelType.STRING)
        name19 = pvproperty(value='name', name='.NM19',
                            dtype=ChannelType.STRING)
        name20 = pvproperty(value='name', name='.NM20',
                            dtype=ChannelType.STRING)
        name21 = pvproperty(value='name', name='.NM21',
                            dtype=ChannelType.STRING)
        name22 = pvproperty(value='name', name='.NM22',
                            dtype=ChannelType.STRING)
        name23 = pvproperty(value='name', name='.NM23',
                            dtype=ChannelType.STRING)
        name24 = pvproperty(value='name', name='.NM24',
                            dtype=ChannelType.STRING)
        name25 = pvproperty(value='name', name='.NM25',
                            dtype=ChannelType.STRING)
        name26 = pvproperty(value='name', name='.NM26',
                            dtype=ChannelType.STRING)
        name27 = pvproperty(value='name', name='.NM27',
                            dtype=ChannelType.STRING)
        name28 = pvproperty(value='name', name='.NM28',
                            dtype=ChannelType.STRING)
        name29 = pvproperty(value='name', name='.NM29',
                            dtype=ChannelType.STRING)
        name30 = pvproperty(value='name', name='.NM30',
                            dtype=ChannelType.STRING)
        name31 = pvproperty(value='name', name='.NM31',
                            dtype=ChannelType.STRING)
        name32 = pvproperty(value='name', name='.NM32',
                            dtype=ChannelType.STRING)

    names = SubGroup(NamesGroup, prefix='')

    time = pvproperty(name='.T', dtype=float)
    freq = pvproperty(name='.FREQ', dtype=float)
    preset_time = pvproperty(name='.TP', dtype=float)
    auto_count_time = pvproperty(name='.TP1', dtype=float)

    class PresetsGroup(PVGroup):
        preset1 = pvproperty(name='.PR1', dtype=int)
        preset2 = pvproperty(name='.PR2', dtype=int)
        preset3 = pvproperty(name='.PR3', dtype=int)
        preset4 = pvproperty(name='.PR4', dtype=int)
        preset5 = pvproperty(name='.PR5', dtype=int)
        preset6 = pvproperty(name='.PR6', dtype=int)
        preset7 = pvproperty(name='.PR7', dtype=int)
        preset8 = pvproperty(name='.PR8', dtype=int)
        preset9 = pvproperty(name='.PR9', dtype=int)
        preset10 = pvproperty(name='.PR10', dtype=int)
        preset11 = pvproperty(name='.PR11', dtype=int)
        preset12 = pvproperty(name='.PR12', dtype=int)
        preset13 = pvproperty(name='.PR13', dtype=int)
        preset14 = pvproperty(name='.PR14', dtype=int)
        preset15 = pvproperty(name='.PR15', dtype=int)
        preset16 = pvproperty(name='.PR16', dtype=int)
        preset17 = pvproperty(name='.PR17', dtype=int)
        preset18 = pvproperty(name='.PR18', dtype=int)
        preset19 = pvproperty(name='.PR19', dtype=int)
        preset20 = pvproperty(name='.PR20', dtype=int)
        preset21 = pvproperty(name='.PR21', dtype=int)
        preset22 = pvproperty(name='.PR22', dtype=int)
        preset23 = pvproperty(name='.PR23', dtype=int)
        preset24 = pvproperty(name='.PR24', dtype=int)
        preset25 = pvproperty(name='.PR25', dtype=int)
        preset26 = pvproperty(name='.PR26', dtype=int)
        preset27 = pvproperty(name='.PR27', dtype=int)
        preset28 = pvproperty(name='.PR28', dtype=int)
        preset29 = pvproperty(name='.PR29', dtype=int)
        preset30 = pvproperty(name='.PR30', dtype=int)
        preset31 = pvproperty(name='.PR31', dtype=int)
        preset32 = pvproperty(name='.PR32', dtype=int)

    presets = SubGroup(PresetsGroup, prefix='')

    class GatesGroup(PVGroup):
        gate1 = pvproperty(name='.G1', dtype=int)
        gate2 = pvproperty(name='.G2', dtype=int)
        gate3 = pvproperty(name='.G3', dtype=int)
        gate4 = pvproperty(name='.G4', dtype=int)
        gate5 = pvproperty(name='.G5', dtype=int)
        gate6 = pvproperty(name='.G6', dtype=int)
        gate7 = pvproperty(name='.G7', dtype=int)
        gate8 = pvproperty(name='.G8', dtype=int)
        gate9 = pvproperty(name='.G9', dtype=int)
        gate10 = pvproperty(name='.G10', dtype=int)
        gate11 = pvproperty(name='.G11', dtype=int)
        gate12 = pvproperty(name='.G12', dtype=int)
        gate13 = pvproperty(name='.G13', dtype=int)
        gate14 = pvproperty(name='.G14', dtype=int)
        gate15 = pvproperty(name='.G15', dtype=int)
        gate16 = pvproperty(name='.G16', dtype=int)
        gate17 = pvproperty(name='.G17', dtype=int)
        gate18 = pvproperty(name='.G18', dtype=int)
        gate19 = pvproperty(name='.G19', dtype=int)
        gate20 = pvproperty(name='.G20', dtype=int)
        gate21 = pvproperty(name='.G21', dtype=int)
        gate22 = pvproperty(name='.G22', dtype=int)
        gate23 = pvproperty(name='.G23', dtype=int)
        gate24 = pvproperty(name='.G24', dtype=int)
        gate25 = pvproperty(name='.G25', dtype=int)
        gate26 = pvproperty(name='.G26', dtype=int)
        gate27 = pvproperty(name='.G27', dtype=int)
        gate28 = pvproperty(name='.G28', dtype=int)
        gate29 = pvproperty(name='.G29', dtype=int)
        gate30 = pvproperty(name='.G30', dtype=int)
        gate31 = pvproperty(name='.G31', dtype=int)
        gate32 = pvproperty(name='.G32', dtype=int)

    gates = SubGroup(GatesGroup, prefix='')

    update_rate = pvproperty(name='.RATE', dtype=int)
    auto_count_update_rate = pvproperty(name='.RAT1', dtype=int)
    egu = pvproperty(value='EGU', name='.EGU', dtype=ChannelType.STRING)


if __name__ == '__main__':
    ioc_options, run_options = ioc_arg_parser(
        default_prefix='scaler_tests:',
        desc="ophyd.tests.test_scaler test IOC")
    ioc = EpicsScalerGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



import logging

from collections import OrderedDict

from .ophydobj import Kind
from .signal import (EpicsSignal, EpicsSignalRO)
from .device import Device
from .device import (Component as Cpt, DynamicDeviceComponent as DDCpt,
                     FormattedComponent as FCpt)

logger = logging.getLogger(__name__)


def _scaler_fields(cls, attr_base, field_base, range_, **kwargs):
    defn = OrderedDict()
    for i in range_:
        attr = '{attr}{i}'.format(attr=attr_base, i=i)
        suffix = '{field}{i}'.format(field=field_base, i=i)
        defn[attr] = (cls, suffix, kwargs)

    return defn


class EpicsScaler(Device):
    '''SynApps Scaler Record interface'''
    # tigger + trigger mode
    count = Cpt(EpicsSignal, '.CNT', trigger_value=1, kind=Kind.omitted)
    count_mode = Cpt(EpicsSignal, '.CONT', string=True, kind=Kind.config)

    # delay from triggering to starting counting
    delay = Cpt(EpicsSignal, '.DLY', kind=Kind.config)
    auto_count_delay = Cpt(EpicsSignal, '.DLY1', kind=Kind.config)

    # the data
    channels = DDCpt(_scaler_fields(EpicsSignalRO, 'chan', '.S', range(1, 33),
                                    kind=Kind.hinted))
    names = DDCpt(_scaler_fields(EpicsSignal, 'name', '.NM', range(1, 33),
                                 kind=Kind.config))

    time = Cpt(EpicsSignal, '.T', kind=Kind.config)
    freq = Cpt(EpicsSignal, '.FREQ', kind=Kind.config)

    preset_time = Cpt(EpicsSignal, '.TP', kind=Kind.config)
    auto_count_time = Cpt(EpicsSignal, '.TP1', kind=Kind.config)

    presets = DDCpt(_scaler_fields(EpicsSignal, 'preset', '.PR', range(1, 33),
                                   kind=Kind.omitted))
    gates = DDCpt(_scaler_fields(EpicsSignal, 'gate', '.G', range(1, 33),
                                 kind=Kind.omitted))

    update_rate = Cpt(EpicsSignal, '.RATE', kind=Kind.config)
    auto_count_update_rate = Cpt(EpicsSignal, '.RAT1', kind=Kind.config)

    egu = Cpt(EpicsSignal, '.EGU', kind=Kind.config)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.stage_sigs.update([('count_mode', 0)])


class ScalerChannel(Device):

    # TODO set up monitor on this to automatically change the name
    chname = FCpt(EpicsSignal, '{self.prefix}.NM{self._ch_num}',
                  kind=Kind.config)
    s = FCpt(EpicsSignalRO, '{self.prefix}.S{self._ch_num}',
             kind=Kind.hinted)
    preset = FCpt(EpicsSignal, '{self.prefix}.PR{self._ch_num}',
                  kind=Kind.config)
    gate = FCpt(EpicsSignal, '{self.prefix}.G{self._ch_num}', string=True,
                kind=Kind.config)

    def __init__(self, prefix, ch_num,
                 **kwargs):
        self._ch_num = ch_num

        super().__init__(prefix, **kwargs)
        self.match_name()

    def match_name(self):
        self.s.name = self.chname.get()


def _sc_chans(attr_fix, id_range):
    defn = OrderedDict()
    for k in id_range:
        defn['{}{:02d}'.format(attr_fix, k)] = (ScalerChannel,
                                                '', {'ch_num': k,
                                                     'kind': Kind.normal})
    return defn


class ScalerCH(Device):

    # The data
    channels = DDCpt(_sc_chans('chan', range(1, 33)))

    # tigger + trigger mode
    count = Cpt(EpicsSignal, '.CNT', trigger_value=1, kind=Kind.omitted)
    count_mode = Cpt(EpicsSignal, '.CONT', string=True, kind=Kind.config)

    # delay from triggering to starting counting
    delay = Cpt(EpicsSignal, '.DLY', kind=Kind.config)
    auto_count_delay = Cpt(EpicsSignal, '.DLY1', kind=Kind.config)

    time = Cpt(EpicsSignal, '.T')
    freq = Cpt(EpicsSignal, '.FREQ', kind=Kind.config)

    preset_time = Cpt(EpicsSignal, '.TP', kind=Kind.config)
    auto_count_time = Cpt(EpicsSignal, '.TP1', kind=Kind.config)

    update_rate = Cpt(EpicsSignal, '.RATE', kind=Kind.omitted)
    auto_count_update_rate = Cpt(EpicsSignal, '.RAT1', kind=Kind.omitted)

    egu = Cpt(EpicsSignal, '.EGU', kind=Kind.config)

    def match_names(self):
        for s in self.channels.component_names:
            getattr(self.channels, s).match_name()

    def select_channels(self, chan_names):
        '''Select channels based on the EPICS name PV

        Parameters
        ----------
        chan_names : Iterable[str] or None

            The names (as reported by the channel.chname signal)
            of the channels to select.
            If *None*, select all channels named in the EPICS scaler.
        '''
        self.match_names()
        name_map = {}
        for s in self.channels.component_names:
            scaler_channel = getattr(self.channels, s)
            nm = scaler_channel.s.name  # as defined in self.match_names()
            if len(nm) > 0:
                name_map[nm] = s

        if chan_names is None:
            chan_names = name_map.keys()

        read_attrs = ['chan01']  # always include time
        for ch in chan_names:
            try:
                read_attrs.append(name_map[ch])
            except KeyError:
                raise RuntimeError("The channel {} is not configured "
                                   "on the scaler.  The named channels are "
                                   "{}".format(ch, tuple(name_map)))
        self.channels.kind = Kind.normal
        self.channels.read_attrs = list(read_attrs)
        self.channels.configuration_attrs = list(read_attrs)
        for ch in read_attrs[1:]:
            getattr(self.channels, ch).s.kind = Kind.hinted



import logging
import pytest
from unittest.mock import Mock

import numpy as np

from ophyd import (Device, Component, FormattedComponent)
from ophyd.signal import (Signal, AttributeSignal, ArrayAttributeSignal,
                          ReadOnlyError)
from ophyd.device import (ComponentWalk, create_device_from_components,
                          required_for_connection, wait_for_lazy_connection,
                          do_not_wait_for_lazy_connection)
from ophyd.utils import ExceptionBundle


logger = logging.getLogger(__name__)


class FakeSignal(Signal):
    def __init__(self, read_pv, *, name=None, parent=None, **kwargs):
        self.read_pv = read_pv
        super().__init__(name=name, parent=parent, **kwargs)
        self._waited_for_connection = False
        self._subscriptions = []

    def wait_for_connection(self):
        self._waited_for_connection = True

    def subscribe(self, method, event_type, **kw):
        self._subscriptions.append((method, event_type, kw))

    def get(self):
        return self.name

    def describe_configuration(self):
        return {self.name + '_conf': {'source': 'SIM:test'}}

    def read_configuration(self):
        return {self.name + '_conf': {'value': 0}}


def test_device_state():
    d = Device('test', name='test')

    d.stage()
    old, new = d.configure({})
    assert old == new
    d.unstage()


def test_attrs():
    class MyDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')

    d = MyDevice('prefix', read_attrs=['cpt1'],
                 configuration_attrs=['cpt2'],
                 name='test'
                 )

    d.read()
    assert d.read_attrs == ['cpt1']
    assert d.configuration_attrs == ['cpt2']

    assert list(d.read().keys()) == [d.cpt1.name]
    assert set(d.read_configuration().keys()) == {d.cpt2.name + '_conf'}

    assert list(d.describe().keys()) == [d.cpt1.name]
    assert set(d.describe_configuration().keys()) == {d.cpt2.name + '_conf', }


def test_complexdevice():
    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')

    class SubSubDevice(SubDevice):
        cpt4 = Component(FakeSignal, '4')

    class MyDevice(Device):
        sub1 = Component(SubDevice, '1')
        subsub2 = Component(SubSubDevice, '2')
        cpt3 = Component(FakeSignal, '3')

    device = MyDevice('prefix', name='dev')
    device.configuration_attrs = ['sub1',
                                  'subsub2.cpt2',
                                  'subsub2.cpt4',
                                  'cpt3']
    device.sub1.read_attrs = ['cpt2']
    device.sub1.configuration_attrs = ['cpt1']

    assert device.sub1.parent is device
    assert device.subsub2.parent is device
    assert device.cpt3.parent is device

    assert device.sub1.component_names == ('cpt1', 'cpt2', 'cpt3')
    assert device.subsub2.component_names == ('cpt1', 'cpt2', 'cpt3', 'cpt4')

    with pytest.raises(AttributeError):
        device.component_names.remove('sub1')

    conf_keys = {'dev_sub1_cpt1_conf',     # from sub1.*
                 # 'dev_sub1_cpt2_conf',   # not in sub1.config_attrs
                 'dev_subsub2_cpt2_conf',  # from subsub2.cpt2
                 'dev_subsub2_cpt4_conf',  # from subsub2.cpt4
                 'dev_cpt3_conf',          # from cpt3


                 }

    assert set(device.describe_configuration().keys()) == conf_keys
    assert set(device.read_configuration().keys()) == conf_keys


def test_complexdevice_stop():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, '4')

        def stop(self, *, success=False):
            self.stop_called = True
            self.success = success
            if self.prefix.endswith('_raises_'):
                raise Exception('stop failed for some reason')

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')
        subsub = Component(SubSubDevice, '')

        def stop(self, *, success=False):
            self.stop_called = True
            self.success = success
            super().stop(success=success)

    class MyDevice(Device):
        sub1 = Component(SubDevice, '1')
        sub2 = Component(SubDevice, '_raises_')
        sub3 = Component(SubDevice, '_raises_')
        cpt3 = Component(FakeSignal, '3')

    dev = MyDevice('', name='mydev')
    with pytest.raises(ExceptionBundle) as cm:
        dev.stop()

    ex = cm.value
    assert len(ex.exceptions) == 2
    assert dev.sub1.stop_called
    assert dev.sub2.stop_called
    assert dev.sub3.stop_called
    assert not dev.sub1.success
    assert not dev.sub2.success
    assert not dev.sub3.success

    assert dev.sub1.subsub.stop_called
    assert dev.sub2.subsub.stop_called
    assert dev.sub3.subsub.stop_called
    assert not dev.sub1.subsub.success
    assert not dev.sub2.subsub.success
    assert not dev.sub3.subsub.success

    dev = MyDevice('', name='mydev')
    with pytest.raises(ExceptionBundle) as cm:
        dev.stop(success=True)

    ex = cm.value
    assert len(ex.exceptions) == 2
    assert dev.sub1.stop_called
    assert dev.sub2.stop_called
    assert dev.sub3.stop_called
    assert dev.sub1.success
    assert dev.sub2.success
    assert dev.sub3.success

    assert dev.sub1.subsub.stop_called
    assert dev.sub2.subsub.stop_called
    assert dev.sub3.subsub.stop_called
    assert dev.sub1.subsub.success
    assert dev.sub2.subsub.success
    assert dev.sub3.subsub.success


def test_name_shadowing():
    RESERVED_ATTRS = ['name', 'parent', 'component_names', '_signals',
                      '_sig_attrs',
                      '_sub_devices']

    type('a', (Device,), {'a': None})  # legal class definition
    # Illegal class definitions:
    for attr in RESERVED_ATTRS:
        with pytest.raises(TypeError):
            type('a', Device, attr=None)


def test_formatted_component():
    FC = FormattedComponent

    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')
        ch = FC(FakeSignal, '{self.prefix}{self._ch}')

        def __init__(self, prefix, ch='a', **kwargs):
            self._ch = ch
            super().__init__(prefix, **kwargs)

    ch_value = '_test_'

    device = MyDevice('prefix:', ch=ch_value, name='test')
    assert device.cpt.parent is device
    assert device.ch.parent is device
    assert device._ch is ch_value
    assert device.ch.read_pv == device.prefix + ch_value
    assert device.cpt.read_pv == device.prefix + MyDevice.cpt.suffix


def test_root():
    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test')
    assert d.cpt.root == d
    assert d.root == d


def test_hidden_component():
    class MyDevice(Device):
        _hidden_sig = Component(FakeSignal, 'suffix')
    d = MyDevice('', name='test')
    assert '_hidden_sig' in d.component_names
    assert not hasattr(d.get(), '_hidden_sig')


def test_attribute_signal():
    init_value = 33

    class SubDevice(Device):
        @property
        def prop(self):
            return init_value

        @prop.setter
        def prop(self, value):
            pass

    class MyDevice(Device):
        sub1 = Component(SubDevice, '1')
        attrsig = Component(AttributeSignal, 'prop')
        sub_attrsig = Component(AttributeSignal, 'sub1.prop',
                                write_access=False)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._value = init_value

        @property
        def prop(self):
            return self._value

        @prop.setter
        def prop(self, value):
            self._value = value

    dev = MyDevice('', name='mydev')
    assert (dev.describe()['mydev_attrsig']['source'] == 'PY:mydev.prop')
    assert (dev.describe()['mydev_sub_attrsig']['source'] == 'PY:mydev.sub1.prop')
    assert dev.attrsig.get() == init_value
    cb = Mock()
    dev.attrsig.subscribe(cb)
    dev.attrsig.put(55)
    assert dev.attrsig.get() == 55
    assert cb.called
    assert dev.sub_attrsig.get() == init_value
    assert dev.sub_attrsig.read_access
    assert not dev.sub_attrsig.write_access

    with pytest.raises(ReadOnlyError):
        dev.sub_attrsig.put(0)


def test_attribute_signal_attributeerror():
    class MyDevice(Device):
        sig1 = Component(AttributeSignal, 'unknownattribute')
        sig2 = Component(AttributeSignal, '__class__.__name')
        sig3 = Component(AttributeSignal, 'unknown.attribute')

    dev = MyDevice('', name='mydev')
    pytest.raises(AttributeError, dev.sig1.get)
    pytest.raises(AttributeError, dev.sig2.get)
    pytest.raises(AttributeError, dev.sig3.get)
    pytest.raises(AttributeError, getattr, dev.sig3, 'base')


def test_shadowing_bs_interface_raises_typeerror():
    class LegalDevice(Device):
        cpt = Component(FakeSignal, 'cpt')

    with pytest.raises(TypeError):
        class IllegalDevice(Device):
            # 'read' shadows the bluesky interface and should not be allowed
            read = Component(FakeSignal, 'read')


def test_array_attribute_signal():
    init_value = range(10)

    class MyDevice(Device):
        attrsig = Component(ArrayAttributeSignal, 'prop')

        @property
        def prop(self):
            return init_value

    dev = MyDevice('', name='mydev')
    np.testing.assert_array_equal(dev.attrsig.get(), init_value)
    assert isinstance(dev.attrsig.get(), np.ndarray)
    assert isinstance(dev.attrsig.get(), np.ndarray)


def test_signal_names():

    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test')
    with pytest.warns(UserWarning):
        signal_names = d.signal_names

    assert signal_names is d.component_names


def test_summary():
    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test')
    # smoke tests
    d.summary()
    d.__str__()
    d.__repr__()


def test_labels():
    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test', labels={'a', 'b'})
    assert d._ophyd_labels_ == {'a', 'b'}


def test_device_put():
    class MyDevice(Device):
        a = Component(Signal)
        b = Component(Signal)

    d = MyDevice('', name='test')
    d.put((1, 2))
    assert d.get() == (1, 2)

    devtuple = d.get_device_tuple()
    d.put(devtuple(a=10, b=12))
    assert d.get() == (10, 12)

    with pytest.raises(ValueError):
        d.put((1, 2, 3))


def test_lazy_wait_for_connect():
    class MyDevice(Device):
        lazy_wait_for_connection = False
        cpt = Component(FakeSignal, 'suffix', lazy=True)

    d = MyDevice('', name='test')
    with wait_for_lazy_connection(d):
        d.cpt

    assert d.cpt._waited_for_connection


def test_lazy_do_not_wait_for_connect():
    class MyDevice(Device):
        lazy_wait_for_connection = True
        cpt = Component(FakeSignal, 'suffix', lazy=True)

    d = MyDevice('', name='test')
    with do_not_wait_for_lazy_connection(d):
        d.cpt

    assert not d.cpt._waited_for_connection


def test_sub_decorator():
    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix', lazy=True)

        @cpt.sub_default
        def default(self, **kw):
            pass

        @cpt.sub_value
        def value(self, **kw):
            pass

        @cpt.sub_meta
        def metadata(self, **kw):
            pass

        @cpt.sub_default
        @cpt.sub_value
        @cpt.sub_meta
        def multi(self, **kw):
            pass

    d = MyDevice('', name='test')

    subs = set(event_type for method, event_type, kw in d.cpt._subscriptions)
    assert subs == {None, 'value', 'meta'}
    assert len(MyDevice.multi._subscriptions) == 3


def test_walk_components():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, '4')

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')
        subsub = Component(SubSubDevice, '')

    class MyDevice(Device):
        sub1 = Component(SubDevice, 'sub1')
        sub2 = Component(SubDevice, 'sub2')
        sub3 = Component(SubDevice, 'sub3')
        cpt3 = Component(FakeSignal, 'cpt3')

    assert list(MyDevice.walk_components()) == [
        ComponentWalk(ancestors=(MyDevice, ),
                      dotted_name='sub1',
                      item=MyDevice.sub1),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub1.cpt1',
                      item=MyDevice.sub1.cls.cpt1),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub1.cpt2',
                      item=MyDevice.sub1.cls.cpt2),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub1.cpt3',
                      item=MyDevice.sub1.cls.cpt3),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub1.subsub',
                      item=MyDevice.sub1.cls.subsub),
        ComponentWalk(ancestors=(MyDevice, SubDevice, SubSubDevice, ),
                      dotted_name='sub1.subsub.cpt4',
                      item=MyDevice.sub1.cls.subsub.cls.cpt4),
        ComponentWalk(ancestors=(MyDevice, ),
                      dotted_name='sub2',
                      item=MyDevice.sub2),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub2.cpt1',
                      item=MyDevice.sub2.cls.cpt1),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub2.cpt2',
                      item=MyDevice.sub2.cls.cpt2),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub2.cpt3',
                      item=MyDevice.sub2.cls.cpt3),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub2.subsub',
                      item=MyDevice.sub2.cls.subsub),
        ComponentWalk(ancestors=(MyDevice, SubDevice, SubSubDevice, ),
                      dotted_name='sub2.subsub.cpt4',
                      item=MyDevice.sub2.cls.subsub.cls.cpt4),
        ComponentWalk(ancestors=(MyDevice, ),
                      dotted_name='sub3',
                      item=MyDevice.sub3),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub3.cpt1',
                      item=MyDevice.sub3.cls.cpt1),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub3.cpt2',
                      item=MyDevice.sub3.cls.cpt2),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub3.cpt3',
                      item=MyDevice.sub3.cls.cpt3),
        ComponentWalk(ancestors=(MyDevice, SubDevice, ),
                      dotted_name='sub3.subsub',
                      item=MyDevice.sub3.cls.subsub),
        ComponentWalk(ancestors=(MyDevice, SubDevice, SubSubDevice, ),
                      dotted_name='sub3.subsub.cpt4',
                      item=MyDevice.sub3.cls.subsub.cls.cpt4),
        ComponentWalk(ancestors=(MyDevice, ),
                      dotted_name='cpt3',
                      item=MyDevice.cpt3),
    ]


@pytest.mark.parametrize('include_lazy', [False, True])
def test_walk_signals(include_lazy):
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, '4', lazy=True)

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')
        subsub = Component(SubSubDevice, '')

    class MyDevice(Device):
        sub1 = Component(SubDevice, 'sub1')
        sub2 = Component(SubDevice, 'sub2')
        sub3 = Component(SubDevice, 'sub3')
        cpt3 = Component(FakeSignal, 'cpt3')

    print(MyDevice.sub1.cls.cpt1)

    dev = MyDevice('', name='mydev')
    walked_list = list(dev.walk_signals(include_lazy=include_lazy))

    dev.summary()
    expected = [
        ComponentWalk(ancestors=(dev, dev.sub1, ),
                      dotted_name='sub1.cpt1',
                      item=dev.sub1.cpt1),
        ComponentWalk(ancestors=(dev, dev.sub1, ),
                      dotted_name='sub1.cpt2',
                      item=dev.sub1.cpt2),
        ComponentWalk(ancestors=(dev, dev.sub1, ),
                      dotted_name='sub1.cpt3',
                      item=dev.sub1.cpt3),
        ComponentWalk(ancestors=(dev, dev.sub1, dev.sub1.subsub, ),
                      dotted_name='sub1.subsub.cpt4',
                      item=dev.sub1.subsub.cpt4),
        ComponentWalk(ancestors=(dev, dev.sub2, ),
                      dotted_name='sub2.cpt1',
                      item=dev.sub2.cpt1),
        ComponentWalk(ancestors=(dev, dev.sub2, ),
                      dotted_name='sub2.cpt2',
                      item=dev.sub2.cpt2),
        ComponentWalk(ancestors=(dev, dev.sub2, ),
                      dotted_name='sub2.cpt3',
                      item=dev.sub2.cpt3),
        ComponentWalk(ancestors=(dev, dev.sub2, dev.sub2.subsub, ),
                      dotted_name='sub2.subsub.cpt4',
                      item=dev.sub2.subsub.cpt4),
        ComponentWalk(ancestors=(dev, dev.sub3, ),
                      dotted_name='sub3.cpt1',
                      item=dev.sub3.cpt1),
        ComponentWalk(ancestors=(dev, dev.sub3, ),
                      dotted_name='sub3.cpt2',
                      item=dev.sub3.cpt2),
        ComponentWalk(ancestors=(dev, dev.sub3, ),
                      dotted_name='sub3.cpt3',
                      item=dev.sub3.cpt3),
        ComponentWalk(ancestors=(dev, dev.sub3, dev.sub3.subsub, ),
                      dotted_name='sub3.subsub.cpt4',
                      item=dev.sub3.subsub.cpt4),
        ComponentWalk(ancestors=(dev, ),
                      dotted_name='cpt3',
                      item=dev.cpt3),
    ]

    if not include_lazy:
        expected = [item for item in expected
                    if 'cpt4' not in item.dotted_name
                    ]

    assert walked_list == expected


def test_walk_subdevice_classes():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, '4')

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')
        subsub = Component(SubSubDevice, '')

    class MyDevice(Device):
        sub1 = Component(SubDevice, 'sub1')
        sub2 = Component(SubDevice, 'sub2')
        sub3 = Component(SubDevice, 'sub3')
        cpt3 = Component(FakeSignal, 'cpt3')

    assert list(MyDevice.walk_subdevice_classes()) == [
        ('sub1', SubDevice),
        ('sub1.subsub', SubSubDevice),
        ('sub2', SubDevice),
        ('sub2.subsub', SubSubDevice),
        ('sub3', SubDevice),
        ('sub3.subsub', SubSubDevice),
    ]


def test_walk_subdevices():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, '4')

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')
        cpt2 = Component(FakeSignal, '2')
        cpt3 = Component(FakeSignal, '3')
        subsub = Component(SubSubDevice, '')

    class MyDevice(Device):
        sub1 = Component(SubDevice, 'sub1')
        sub2 = Component(SubDevice, 'sub2')
        sub3 = Component(SubDevice, 'sub3')
        cpt3 = Component(FakeSignal, 'cpt3')

    dev = MyDevice('', name='mydev')
    assert list(dev.walk_subdevices()) == [
        ('sub1', dev.sub1),
        ('sub1.subsub', dev.sub1.subsub),
        ('sub2', dev.sub2),
        ('sub2.subsub', dev.sub2.subsub),
        ('sub3', dev.sub3),
        ('sub3.subsub', dev.sub3.subsub),
    ]


def test_dotted_name():
    from ophyd import Device, Component as Cpt
    from ophyd.sim import SynSignal

    class Inner(Device):
        x = Cpt(SynSignal)
        y = Cpt(SynSignal)

    class Outer(Device):
        a = Cpt(Inner)
        b = Cpt(Inner)

    o = Outer(name='test')

    assert o.dotted_name == ''
    assert o.a.dotted_name == 'a'
    assert o.b.dotted_name == 'b'

    assert o.a.x.dotted_name == 'a.x'
    assert o.b.x.dotted_name == 'b.x'

    assert o.a.y.dotted_name == 'a.y'
    assert o.b.y.dotted_name == 'b.y'

    assert o.attr_name == ''
    assert o.a.attr_name == 'a'
    assert o.b.attr_name == 'b'

    assert o.a.x.attr_name == 'x'
    assert o.b.x.attr_name == 'x'

    assert o.a.y.attr_name == 'y'
    assert o.b.y.attr_name == 'y'


def test_create_device():
    components = dict(cpt1=Component(Signal, value=0),
                      cpt2=Component(Signal, value=1),
                      cpt3=Component(Signal, value=2)
                      )
    Dev = create_device_from_components('Dev', base_class=Device,
                                        **components
                                        )
    assert Dev.__name__ == 'Dev'
    assert Dev.cpt1 is components['cpt1']
    dev = Dev(name='dev')
    assert dev.cpt1.get() == 0
    assert dev.cpt2.get() == 1
    assert dev.cpt3.get() == 2


def test_create_device_bad_component():
    with pytest.raises(ValueError):
        create_device_from_components('Dev', base_class=Device,
                                      bad_component=None)


def test_required_for_connection_on_method_with_subscriptions():
    class MyDevice(Device):
        cpt = Component(Signal, value=0)

        @required_for_connection
        @cpt.sub_value
        def method(self):
            ...

    dev = MyDevice(name='dev')

    with pytest.raises(TimeoutError):
        dev.wait_for_connection(timeout=0.1)

    dev.cpt.put(0)
    dev.wait_for_connection(timeout=0.1)


def test_required_for_connection_on_method():
    class MyDevice(Device):
        @required_for_connection
        def method(self):
            ...

    dev = MyDevice(name='dev')

    # Timeout without it having been called:
    with pytest.raises(TimeoutError):
        dev.wait_for_connection(timeout=0.01)

    # Call and expect no timeout:
    dev.method()
    dev.wait_for_connection(timeout=0.01)


def test_required_for_connection_in_init():
    class MyDevice(Device):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.call_to_connect = required_for_connection(self.method,
                                                           device=self)

        def method(self):
            print('method called')

    dev = MyDevice(name='dev')

    # Timeout without it having been called:
    with pytest.raises(TimeoutError):
        dev.wait_for_connection(timeout=0.01)

    # Call and expect no timeout:
    dev.call_to_connect()
    dev.wait_for_connection(timeout=0.01)


def test_noneified_component():
    class SubDevice(Device):
        cpt1 = Component(FakeSignal, '1')

    class MyDevice(Device):
        sub1 = Component(SubDevice, 'sub1')
        sub2 = Component(SubDevice, 'sub2')

    class MyDeviceWithoutSub2(Device):
        sub1 = Component(SubDevice, 'sub1')
        sub2 = None

    assert MyDevice.component_names == ('sub1', 'sub2')
    assert MyDevice._sub_devices == ['sub1', 'sub2']

    assert MyDeviceWithoutSub2.component_names == ('sub1', )
    assert MyDeviceWithoutSub2._sub_devices == ['sub1']


@pytest.mark.parametrize('lazy_state, cntx',
                         [(False, wait_for_lazy_connection),
                          (True, do_not_wait_for_lazy_connection)])
def test_lazy_wait_context(lazy_state, cntx):
    class LocalExcepton(Exception):
        ...

    d = Device(name='d')
    d.lazy_wait_for_connection = lazy_state

    try:
        with cntx(d):
            assert d.lazy_wait_for_connection == (not lazy_state)
            raise LocalExcepton
    except LocalExcepton:
        ...

    assert d.lazy_wait_for_connection is lazy_state



# vi: ts=4 sw=4
'''AreaDetector Devices

`areaDetector`_ detector abstractions

.. _areaDetector: http://cars.uchicago.edu/software/epics/areaDetector.html
'''

from .base import (ADBase, ADComponent as C)
from . import cam


__all__ = ['DetectorBase',
           'AreaDetector',
           'AdscDetector',
           'Andor3Detector',
           'AndorDetector',
           'BrukerDetector',
           'DexelaDetector',
           'FirewireLinDetector',
           'FirewireWinDetector',
           'GreatEyesDetector',
           'LightFieldDetector',
           'Mar345Detector',
           'MarCCDDetector',
           'PSLDetector',
           'PerkinElmerDetector',
           'PilatusDetector',
           'PixiradDetector',
           'PointGreyDetector',
           'ProsilicaDetector',
           'PvcamDetector',
           'RoperDetector',
           'SimDetector',
           'URLDetector',
           ]


class DetectorBase(ADBase):
    """
    The base class for the hardware-specific classes that follow.

    Note that Plugin also inherits from ADBase.
    This adds some AD-specific methods that are not shared by the plugins.
    """
    _default_configuration_attrs = (ADBase._default_configuration_attrs +
                                    ('cam', ))

    def dispatch(self, key, timestamp):
        """Notify plugins of acquisition being complete.

        When a new acquisition is started, this method is called with a
        key which is a label like 'light', 'dark', or 'gain8'.

        It in turn calls ``generate_datum`` on all of the plugins that have
        that method.

        File plugins are identified by searching for a
        :meth:`~ophyd.areadetector.filestore_mixins.FileStoreBase.generate_datum`
        method that must have the signature ::

           def generate_datum(key: str, timestamp: float, datum_kwargs: dict):
              ...

        """
        file_plugins = [s for s in self._signals.values() if
                        hasattr(s, 'generate_datum')]
        for p in file_plugins:
            if p.enable.get():
                p.generate_datum(key, timestamp, {})

    def make_data_key(self):
        source = 'PV:{}'.format(self.prefix)
        shape = tuple(self.cam.array_size.get())
        return dict(shape=shape, source=source, dtype='array',
                    external='FILESTORE:')

    def collect_asset_docs(self):
        file_plugins = [s for s in self._signals.values() if
                        hasattr(s, 'collect_asset_docs')]
        for p in file_plugins:
            yield from p.collect_asset_docs()


class AreaDetector(DetectorBase):
    cam = C(cam.AreaDetectorCam, 'cam1:')


class SimDetector(DetectorBase):
    _html_docs = ['simDetectorDoc.html']
    cam = C(cam.SimDetectorCam, 'cam1:')


class AdscDetector(DetectorBase):
    _html_docs = ['adscDoc.html']
    cam = C(cam.AdscDetectorCam, 'cam1:')


class AndorDetector(DetectorBase):
    _html_docs = ['andorDoc.html']
    cam = C(cam.AndorDetectorCam, 'cam1:')


class Andor3Detector(DetectorBase):
    _html_docs = ['andor3Doc.html']
    cam = C(cam.Andor3DetectorCam, 'cam1:')


class BrukerDetector(DetectorBase):
    _html_docs = ['BrukerDoc.html']
    cam = C(cam.BrukerDetectorCam, 'cam1:')


class DexelaDetector(DetectorBase):
    _html_docs = ['DexelaDoc.html']
    cam = C(cam.DexelaDetectorCam, 'cam1:')


class FirewireLinDetector(DetectorBase):
    _html_docs = ['FirewireWinDoc.html']
    cam = C(cam.FirewireLinDetectorCam, 'cam1:')


class FirewireWinDetector(DetectorBase):
    _html_docs = ['FirewireWinDoc.html']
    cam = C(cam.FirewireWinDetectorCam, 'cam1:')


class GreatEyesDetector(DetectorBase):
    _html_docs = []  # the documentation is not public
    cam = C(cam.GreatEyesDetectorCam, 'cam1:')


class LightFieldDetector(DetectorBase):
    _html_docs = ['LightFieldDoc.html']
    cam = C(cam.LightFieldDetectorCam, 'cam1:')


class Mar345Detector(DetectorBase):
    _html_docs = ['Mar345Doc.html']
    cam = C(cam.Mar345DetectorCam, 'cam1:')


class MarCCDDetector(DetectorBase):
    _html_docs = ['MarCCDDoc.html']
    cam = C(cam.MarCCDDetectorCam, 'cam1:')


class PerkinElmerDetector(DetectorBase):
    _html_docs = ['PerkinElmerDoc.html']
    cam = C(cam.PerkinElmerDetectorCam, 'cam1:')


class PSLDetector(DetectorBase):
    _html_docs = ['PSLDoc.html']
    cam = C(cam.PSLDetectorCam, 'cam1:')


class PilatusDetector(DetectorBase):
    _html_docs = ['pilatusDoc.html']
    cam = C(cam.PilatusDetectorCam, 'cam1:')


class PixiradDetector(DetectorBase):
    _html_docs = ['PixiradDoc.html']
    cam = C(cam.PixiradDetectorCam, 'cam1:')


class PointGreyDetector(DetectorBase):
    _html_docs = ['PointGreyDoc.html']
    cam = C(cam.PointGreyDetectorCam, 'cam1:')


class ProsilicaDetector(DetectorBase):
    _html_docs = ['prosilicaDoc.html']
    cam = C(cam.ProsilicaDetectorCam, 'cam1:')


class PvcamDetector(DetectorBase):
    _html_docs = ['pvcamDoc.html']
    cam = C(cam.PvcamDetectorCam, 'cam1:')


class RoperDetector(DetectorBase):
    _html_docs = ['RoperDoc.html']
    cam = C(cam.RoperDetectorCam, 'cam1:')


class URLDetector(DetectorBase):
    _html_docs = ['URLDoc.html']
    cam = C(cam.URLDetectorCam, 'cam1:')



#!/usr/bin/env python3
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run


class SignalTestIOC(PVGroup):
    read_only = pvproperty(value=0.0, read_only=True, alarm_group='alarm_a')
    read_write = pvproperty(value=0.0,
                            lower_ctrl_limit=-100.0,
                            upper_ctrl_limit=100.0,
                            alarm_group='alarm_a')

    pair_rbv = pvproperty(value=0.0, read_only=True, alarm_group='alarm_a')
    pair_set = pvproperty(value=0.0,
                          lower_ctrl_limit=-100.0,
                          upper_ctrl_limit=100.0,
                          alarm_group='alarm_a')

    @pair_set.putter
    async def pair_set(self, instance, value):
        await self.pair_rbv.write(value=value)

    waveform = pvproperty(value=[ord('a'), ord('b'), ord('c')], read_only=True,
                          alarm_group='alarm_a')
    bool_enum = pvproperty(value=True, alarm_group='alarm_a')
    alarm_status = pvproperty(value=0)
    set_severity = pvproperty(value=0)

    @set_severity.putter
    async def set_severity(self, instance, severity):
        await self.read_only.alarm.write(
            severity=severity, status=self.alarm_status.value)


if __name__ == '__main__':
    ioc_options, run_options = ioc_arg_parser(
        default_prefix='signal_tests:',
        desc="ophyd.tests.test_signal test IOC")
    ioc = SignalTestIOC(**ioc_options)
    run(ioc.pvdb, **run_options)



import pytest
import os
import logging
import numpy as np
import tempfile

from ophyd.utils import epics_pvs as epics_utils
from ophyd.utils import (make_dir_tree, makedirs, set_and_wait)
from ophyd import Signal


logger = logging.getLogger(__name__)


def test_split():
    utils = epics_utils

    assert utils.split_record_field('record.field') == ('record', 'field')
    assert utils.split_record_field('record.field.invalid') == ('record.field',
                                                                'invalid')
    assert utils.strip_field('record.field') == 'record'
    assert utils.strip_field('record.field.invalid') == 'record.field'
    assert utils.record_field('record', 'field') == 'record.FIELD'


def test_waveform_to_string():
    s = 'abcdefg'
    asc = [ord(c) for c in s]
    assert epics_utils.waveform_to_string(asc) == s

    asc = [ord(c) for c in s] + [0, 0, 0]
    assert epics_utils.waveform_to_string(asc) == s


def test_records_from_db():
    # db_dir = os.path.join(config.epics_base, 'db')

    # if os.path.exists(db_dir):
    #     # fall back on the db file included with the tests
    db_dir = os.path.dirname(__file__)
    db_path = os.path.join(db_dir, 'scaler.db')
    records = epics_utils.records_from_db(db_path)
    assert ('bo', '$(P)$(S)_calcEnable') in records


@pytest.mark.parametrize('value, dtype, shape', [
    [1, 'integer', []],
    [1.0, 'number', []],
    [1e-3, 'number', []],
    ['foo', 'string', []],
    [np.array([1, 2, 3]), 'array', [3]],
    [np.array([[1, 2], [3, 4]]), 'array', [2, 2]],
    [(1, 2, 3), 'array', [3]],
    [[1, 2, 3], 'array', [3]],
    [[], 'array', [0]]
]
)
def test_data_type_and_shape(value, dtype, shape):
    utils = epics_utils
    assert utils.data_type(value) == dtype
    assert utils.data_shape(value) == shape


@pytest.mark.parametrize('value',
                         [dict()])
def test_invalid_data_type(value):
    utils = epics_utils
    with pytest.raises(ValueError):
        utils.data_type(value)


def assert_OD_equal_ignore_ts(a, b):
    for (k1, v1), (k2, v2) in zip(a.items(), b.items()):
        assert (k1 == k2) and (v1['value'] == v2['value'])


def assert_file_mode(path, expected):
    assert (os.stat(path).st_mode & 0o777) == expected


def test_makedirs():
    with tempfile.TemporaryDirectory() as tempdir:
        create_dir = os.path.join(tempdir, 'a')
        makedirs(create_dir, mode=0o767, mode_base=tempdir)
        assert_file_mode(create_dir, 0o767)


def test_make_dir_tree():
    with tempfile.TemporaryDirectory() as tempdir:
        paths = make_dir_tree(2016, base_path=tempdir, mode=0o777)
        assert len(paths) == 366

        for path in paths:
            assert_file_mode(path, 0o777)

        assert os.path.join(tempdir, '2016', '03', '04') in paths
        assert os.path.join(tempdir, '2016', '02', '29') in paths


def test_valid_pvname():
    with pytest.raises(epics_utils.BadPVName):
        epics_utils.validate_pv_name('this.will.fail')


def test_array_into_softsignal():
    data = np.array([1, 2, 3])
    s = Signal(name='np.array')
    set_and_wait(s, data)
    assert np.all(s.get() == data)
